import math
import numpy as np
import pandas as pd


def grey_relation(df, rho=0.5, func=0, mother=''):
    '''   
    Func:灰色关联\n
    IN:dataframe原始数据\n
    OUT:series每个样本的评分\n
    rhp --> 关联度一般为0.5\n
    func --> 0：回归；1：评价\n
    mother --> 母序列，默认为空
    '''
    df = df/df.mean()
    df_temp = df.copy(deep=True)
    if func == 1:
        # 评价
        mother_series = df.max(axis=1)
    else:
        # 回归
        mother_series = df[mother]
        df = df.drop(mother, axis=1)
    df = np.abs(df.sub(mother_series, axis=0))
    min = df.min().min()
    max = df.max().max()
    grey = (min+rho*max)/(df+rho*max)
    weights = grey.sum()/df.index.size
    if func == 0:
        # 回归
        return weights
    else:
        # 评价
        weights = weights/weights.sum()
        return df_temp.dot(weights)


def entropy_weight(df):
    '''
    Func:熵权法计算权重\n
    In:dataframe各指标数据\n
    Out:series各指标权重
    '''
    # 归一化
    df = df.apply(lambda x: ((x - np.min(x)) / (np.max(x) - np.min(x))))
    # 求k
    k = 1 / math.log(df.index.size)
    # 信息熵
    df_p = df/df.sum()
    df_inf = -np.log(df_p).replace([np.inf, -np.inf], 0)*df_p*k
    # 计算冗余度
    redundancy = 1 - df_inf.sum()
    # 计算各指标的权重
    weights = redundancy/redundancy.sum()
    return weights


def topsis(df, weights=[]):
    '''
    Func:TOPSIS法\n
    IN:dataframe各指标数据\n
    OUT:series每个样本的分数\n
    weights --> 按顺序传入指标权重
    '''
    df = df/((df**2).sum())
    series_max = df.max()
    series_min = df.min()
    if weights == []:
        weights = np.ones(df.columns.size)/df.columns.size
    D_max = ((weights*((df-series_max)**2)).sum(axis=1))**(1/2)
    D_min = ((weights*((df-series_min)**2)).sum(axis=1))**(1/2)
    D = D_min/(D_min+D_max)
    return D/D.sum()


def entropy_weight_topsis(df):
    '''
    Func:基于熵权法修正的TOPSIS法\n
    IN:dataframe各指标数据\n
    OUT:series每个样本的分数
    '''
    return topsis(df, list((entropy_weight(df))))


def ladder_distribution(index_dict={'A': [1, 2, 3], 'B': [2, 3, 4], 'C': [3, 4, 5]}, score_dict={'A': 2.9, 'B': 5, 'C': 4.1}):
    '''
    Func:梯型分布模糊综合评价\n
    OUT:模糊综合评价矩阵R
    '''
    def calculate(index_list, score):
        result = [0]*len(index_list)
        index_list.append(score)
        index_list.sort()
        index = index_list.index(score)
        if index > 0 and index < len(index_list)-1:
            left = index_list[index-1]
            right = index_list[index+1]
            result[index] = (score-left)/(right-left)
            result[index-1] = (right-score)/(right-left)
        elif index == 0:
            result[0] = 1
        else:
            result[-1] = 1
        return result
    keys = index_dict.keys()
    retult = [calculate(i[1][0], i[1][1]) for i in enumerate(
        zip(index_dict.values(), score_dict.values()))]
    return pd.DataFrame(dict(zip(keys, retult))).T


def fuzzy_synthesis(weights, R):
    '''
    Func:模糊综合评级\n
    OUT:总隶属度
    weights --> dict,指标权重\n
    R --> 模糊综合评价矩阵R
    '''
    return pd.Series(weights).dot(R)


if __name__ == '__main__':
    test_dict = {'A': [9, 3, 2, 4], 'B': [
        3, 5, 7, 8], 'C': [10, 4, 8, 9], }
    test_df = pd.DataFrame(test_dict)
    t = entropy_weight_topsis(test_df)
    print(t)



import math
import os
import random
import sys
import time
import itertools
from itertools import product
from tqdm import tqdm

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pretty_errors
import scipy.optimize as sco
from tqdm import tqdm, trange

sys.path.append(os.path.dirname(__file__) + os.sep + '../')
try:
    from ..log.log import slog, sprint, hide, show, progress_bar
    from ..data.Stock import StockData
except:
    from log.log import slog, sprint, hide, show, progress_bar
    from data.Stock import StockData


class Markovitz(object):
    '''
    组合投资权重\n
    names=['贵州茅台', '隆基股份', '五粮液']\n
    start_date='2019-01-01'\n
    end_date='2020-01-01'\n
    no_risk_rate=0.0185\n
    funds=10000000
    '''

    def __init__(self, names=['贵州茅台', '隆基股份', '五粮液'],
                 start_date='2019-01-01',
                 end_date='2020-01-01',
                 no_risk_rate=0.0185,
                 funds=10000000):
        self.names = names
        self.lens = len(names)
        self.start_date = start_date
        self.end_date = end_date
        self.no_risk_rate = no_risk_rate
        self.funds = funds
        sprint('Please make sure that all of the stocks are in the market!')
        sprint('Initializing...')
        global StockData
        sd = StockData(names=self.names, start_date=self.start_date,
                       end_date=self.end_date)
        self.datas = sd.stocks_data()
        self.datas.index = self.datas['name']
        self.data = self.datas
        self.date = list(map(lambda x: str(x)[:10], self.data.date.unique()))
        self.first_date = self.date[0]
        self.last_date = self.date[-1]
        self.first_price = self.data[self.data.date == self.data.date.unique(
        )[0]][['open', 'name']].set_index('name').to_dict()['open']
        self.last_price = self.data[self.data.date == self.data.date.unique(
        )[-1]][['close', 'name']].set_index('name').to_dict()['close']
        self.flag = True

    def weights(self, number=5000):
        # 股票随机权重
        return np.random.dirichlet(np.ones(self.lens), size=number)

    def calculate(self):
        data = self.data[['date', 'name', 'pctChg']]
        # 收益率均值
        data_mean = data.groupby('name').mean().T
        # 收益率协方差矩阵
        data_cov = pd.DataFrame()
        for name in self.names:
            data_cov[name] = list(data[data['name'] == name]['pctChg'])
        data_cov = data_cov.cov()
        return data_mean, data_cov

    def scatter_data(self, number=5000):
        '''散点数据，默认生成5000个
        '''
        data_mean, data_cov = self.calculate()
        # 随机权重
        weights = self.weights(number=number)
        # 收益率
        rate = data_mean.dot(weights.T).T
        # 方差
        risk = np.sqrt(np.diagonal(weights.dot(data_cov).dot(weights.T)))
        # 散点
        df_scatter = pd.DataFrame()
        df_scatter['weights'] = pd.Series(map(lambda x: str(x), weights))
        df_scatter['risk'] = risk
        df_scatter['rate'] = rate['pctChg']
        # 夏普比率
        df_scatter['sharpe'] = (
            df_scatter.rate-self.no_risk_rate)/df_scatter.risk
        return df_scatter

    def max_sharpe(self, weights):
        '''
        计算夏普比率
        '''
        data_mean, data_cov = self.calculate()
        weights = np.array(weights)
        rate = data_mean.dot(weights.T)['pctChg']
        risk = np.sqrt(weights.dot(data_cov).dot(weights.T))
        return (self.no_risk_rate-rate)/risk  # 相反数

    def shares_numbers(self):
        '''
        股票手数（不取整）
        '''
        df = pd.DataFrame(columns=['price', 'weight'])
        df['weight'] = pd.Series(self.optimization()[1])
        df['price'] = pd.Series(self.last_price)
        df['shares'] = (self.funds*df['weight'])/(df['price']*100)
        self.shares_number = df
        return df

    def optimization(self):
        '''
        非线性规划找出最大夏普比率
        '''
        opts = sco.minimize(fun=self.max_sharpe,
                            x0=np.ones(self.lens)/self.lens,
                            bounds=tuple((0, 1)for x in range(self.lens)),
                            constraints={'type': 'eq',
                                         'fun': lambda x: np.sum(x) - 1}
                            )
        return -opts.fun, dict(zip(self.names, list(opts.x)))

    def adjacent_combination(self, shares):
        ceils = [math.ceil(i) for i in shares.values()]
        floors = [math.floor(i) for i in shares.values()]
        result = np.array(list(zip(ceils, floors)))
        df_shares = pd.DataFrame(list(product(*result)), columns=self.names)
        return df_shares

    def search(self):
        df = self.shares_numbers()[['price', 'shares']]
        weights = df['shares']*(100*df['price'])/self.funds
        max_sharpe = -self.max_sharpe(weights=weights)
        df_shares = self.adjacent_combination(shares=df['shares'].to_dict()).T
        df_search = df_shares.T
        sharpe_list = []
        for i in range(0, 2**self.lens):
            weights = df_shares[i]*(100*df['price'])/self.funds
            sharpe_list.append(-self.max_sharpe(weights=weights))
        df_search['sharpe'] = sharpe_list
        df_search['<max_sharpe'] = df_search['sharpe'] <= max_sharpe
        df_search = df_search.sort_values(by='sharpe', ascending=False)
        self.original_max_sharpe = max_sharpe
        return df_search

    def exam_search(self, exam_shares):
        df = self.shares_number[['price', 'shares']]
        weights = exam_shares*(100*df['price'])/self.funds
        exam_max_sharpe = -self.max_sharpe(weights=weights)
        exam_shares_list = list(exam_shares)
        exam_lists = []
        for i in range(0, self.lens):
            lists = exam_shares_list.copy()
            lists[i] += 1
            exam_lists.append(lists)
        df_exam = pd.DataFrame(exam_lists, columns=self.names)
        sharpe_list = []
        for exam in exam_lists:
            weights = exam*(100*df['price'])/self.funds
            sharpe_list.append(-self.max_sharpe(weights=weights))
        df_exam['sharpe'] = sharpe_list
        df_exam['cost'] = 0
        for i in df[['price']].itertuples():
            name = i.Index
            price = i.price
            df_exam[f'{name}_cost'] = (df_exam[name]*price*100).astype(int)
            df_exam['cost'] += df_exam[f'{name}_cost']
        df_exam['>=exam_max_sharpe'] = df_exam['sharpe'] >= exam_max_sharpe
        df_exam['<=max_sharpe'] = df_exam['sharpe'] <= self.original_max_sharpe
        df_exam['cost<funds'] = df_exam['cost'] <= self.funds
        df_exam['all'] = df_exam['>=exam_max_sharpe'] & \
            df_exam['<=max_sharpe'] & df_exam['cost<funds']
        df_exam = df_exam.sort_values(by='sharpe', ascending=False)
        return df_exam

    def exam_shares_cycle(self):
        df_search = self.search()
        exam_shares = df_search[df_search['<max_sharpe']
                                == True][self.names]
        lists = []
        for i, j in exam_shares.iterrows():
            exam_share = self.exam_search(exam_shares=j)
            exam_share = exam_share[exam_share['all']
                                    == True][self.names]
            lists.append(exam_share)
        return pd.concat(lists)

    def circle(self, exam_shares):
        exam_list = []
        for i, exam_share in exam_shares.iterrows():
            result = self.exam_search(exam_shares=exam_share)
            exam_list.append(result)
        if not exam_list:
            self.flag = False
            return pd.DataFrame(columns=self.names)
        df = pd.concat(exam_list)
        df = df.drop_duplicates()
        df = df[df['all'] == True]
        if df.empty:
            self.flag = False
        return df

    def tree(self):
        '''分支检验
        '''
        results = []
        sprint('Calculate the optimal number of stocks...')
        for exam_shares in progress_bar(self.exam_shares_cycle()):
            df_exam = self.exam_search(exam_shares=exam_shares)
            exam_shares = df_exam[df_exam['all'] == True][self.names]
            while True:
                exam_shares_last = exam_shares
                print(exam_shares_last)
                exam_shares = self.circle(exam_shares=exam_shares[self.names])
                if not self.flag:
                    results.append(exam_shares_last)
                    break
            self.flag = True
        return pd.concat(results).sort_values(by='sharpe', ascending=False).iloc[0]

    def optimal_weight(self):
        '''
        计算最优权重
        '''
        sharpe_value, weights_dict = self.optimization()
        weights_dict['sharpe'] = sharpe_value
        return pd.Series(weights_dict)

    def buy(self):
        '''股票购买数量（整数）
        '''
        items = self.names+['sharpe', 'cost'] + \
            list(map(lambda x: x+'_cost', self.names))
        return self.tree()[items]

    def font_scatter_data(self, number=500):
        '''边界散点数据，默认生成500个
        '''
        scatter_data = self.scatter_data()
        data_mean, data_cov = self.calculate()
        scatter_list = []
        sprint('Searching for boundary scatter...')
        for i in trange(number):
            i
            random_rate = random.uniform(
                scatter_data.rate.min(), scatter_data.rate.max())
            constraints = ({'type': 'eq', 'fun': lambda weights: weights.sum()-1},
                           {'type': 'eq', 'fun': lambda weights: data_mean.dot(weights.T)['pctChg']-random_rate})
            opts = sco.minimize(fun=lambda weights: weights.dot(data_cov).dot(weights.T),
                                x0=np.ones(self.lens)/self.lens,
                                bounds=tuple((0, 1)for x in range(self.lens)),
                                constraints=constraints
                                )
            scatter_list.append([opts.x,  np.sqrt(opts.fun), random_rate])
        scatter_df = pd.DataFrame(scatter_list, columns=[
                                  'weights', 'risk', 'rate'])
        scatter_df['sharpe'] = (
            scatter_df.rate-self.no_risk_rate)/scatter_df.risk
        return scatter_df

    def drawing(self):
        '''散点图
        '''
        max_sharpe = self.optimization()[0]
        scatter_data = self.scatter_data(number=10000)
        font_scatter_data = self.font_scatter_data(number=1000)
        sprint(f'max sharpe: {max_sharpe}')
        plt.style.use('seaborn-paper')
        plt.scatter(scatter_data.risk, scatter_data.rate,
                    s=10, marker=".", c='b')
        plt.scatter(font_scatter_data.risk, font_scatter_data.rate,
                    s=10, marker=".", c='r')
        plt.axline(xy1=(0, self.no_risk_rate), slope=max_sharpe, c='m')
        plt.xlabel('Risk')
        plt.ylabel('Yield')
        plt.show()

    def weights_test(self, stocks=['药明康德', '比亚迪', '山西汾酒', '五粮液', '贵州茅台'], number=5):

        self.names = stocks
        self.lens = len(self.names)
        global StockData
        self.data = self.datas.loc[self.names]
        self.data = self.data.reset_index(drop=True)
        return self.optimal_weight()

    def weights_tests(self, stocks=['药明康德', '比亚迪', '山西汾酒', '五粮液', '贵州茅台'], number=5):
        lists = []
        for i in tqdm(list(itertools.combinations(stocks, number))):
            result = self.weights_test(list(i), number)
            result_dict = result.to_dict()
            sharpe = result_dict['sharpe']
            del result_dict['sharpe']
            std = np.std(list(result_dict.values()))
            lists.append(
                [result_dict, sharpe, std, np.min(list(result_dict.values()))])
            if False not in list(result > 0.02)[:-1]:
                print('--'*30)
                print(result.map(lambda x: round(x, 3)))
                print('--'*30)
        return pd.DataFrame(lists, columns=['weights', 'sharpe', 'min weight'])


if __name__ == '__main__':
    stocks = [
        '贵州茅台', '泸州老窖', '迎驾贡酒',
        '明微电子', '洁美科技', '晶丰明源',
        '海康威视', '柏楚电子', '恒生电子',
        '老板电器', '火星人', '科沃斯',
        '凯普生物', '一品红', '健民集团',
        '西藏珠峰', '西部超导', '天华超净'
    ]
    mk = Markovitz(names=stocks,  # 股票组合
                   start_date='2021-05-04',  # 开始日期
                   end_date='2021-11-04',  # 结束日期
                   no_risk_rate=0.023467,  # 无风险利率
                   funds=10000000  # 最大资金限制
                   )
    mk.weights_tests(stocks=stocks, number=5).to_excel(
        'result.xlsx', index=False)



import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats import kstest
import matplotlib
import matplotlib.pyplot as plt
from itertools import combinations
from scipy.interpolate import interp1d
matplotlib.rcParams['font.sans-serif'] = ['SimHei']   # 用黑体显示中文
matplotlib.rcParams['axes.unicode_minus'] = False     # 正常显示负号


class DataPreprocessor(object):
    '''
    Func:数据预处理\n
    df --> 原始数据
    '''

    def __init__(self, df):
        self.df = df
        self.cubic_dict = {}

    def normalize(self, columns=[]):
        '''
        Func:归一化选定变量
        '''
        for column in columns:
            max = self.df[column].max()
            min = self.df[column].min()
            self.df[column] = self.df[column].map(
                lambda x: (x - min) / (max - min))

    def normalize_all(self):
        '''
        Func:归一化全部变量
        '''
        self.df = self.df.apply(lambda x: (
            (x - np.min(x)) / (np.max(x) - np.min(x))))

    def direction(self, column, type=1, best_value=0, range=(0, 1)):
        '''
        type=0 --> 极大型，越大越好\n
        type=1 --> 极小型，越小越好\n
        type=2 --> 中间型，接近某个值越好，需传入best_value\n
        type=3 --> 区域型，落在某个区间越好，需传入range
        '''
        def rgnl(x):
            min_value = range[0]
            max_value = range[1]
            M = max(min_value-np.min(self.df[column]),
                    np.max(self.df[column])-max_value)
            if x < min_value:
                return 1-(min_value-x)/M
            elif x > max_value:
                return 1-(x-max_value)/M
            else:
                return 1
        if type == 1:
            self.df[column] = np.max(self.df[column])-self.df[column]
        elif type == 2:
            self.df[column] = 1-abs(self.df[column]-best_value) / \
                np.max(abs(self.df[column]-best_value))
        elif type == 3:
            self.df[column] = self.df[column].map(lambda x: rgnl(x))

    def dummy(self, columns=[]):
        '''
        Func:独热编码选定变量
        '''
        for column in columns:
            self.df = self.df.join(pd.get_dummies(
                self.df[column], prefix=column))
            del self.df[column]

    def dummy_all(self):
        '''
        Func:独热编码全部分类变量
        '''
        self.df = pd.get_dummies(self.df)

    def ks_test(self, columns=[]):
        '''
        Func:检验变量是否符合正态分布
        '''
        df_ks = self.df[columns]
        df_ks.dropna(inplace=True)
        return {column: kstest(df_ks.astype(float), 'norm') for column in columns}

    def correlation(self, columns=[], method='pearson'):
        '''
        Func:计算相关系数\n
        method --> pearson or spearman or kendall
        '''
        df_cor = self.df[columns]
        df_cor.dropna(inplace=True)
        lists = []
        for i in combinations(columns, 2):
            if method == 'pearson':
                result = stats.pearsonr(df_cor[i[0]].astype(
                    float), df_cor[i[1]].astype(float))
            elif method == 'spearman':
                result = stats.spearmanr(df_cor[i[0]].astype(
                    float), df_cor[i[1]].astype(float))
            elif method == 'kendall':
                result = stats.kendalltau(df_cor[i[0]].astype(
                    float), df_cor[i[1]].astype(float))
            else:
                raise ValueError('method is wrong!')
            lists.append([i, result[0], result[1], result[1] < 0.05])
        df = pd.DataFrame(
            lists, columns=['combinations', 'correlation', 'pvalue', 'rejection'])
        return df.set_index('combinations')

    def cubic(self, x, y, kind='cubic'):
        '''
        Func: 插值法(cubic/linear)
        x --> 插值依据序列
        y --> 待插值序列
        '''
        df_null = self.df[[x, y]].isnull()
        df_x = self.df[x][df_null[x] == False]
        df_y = self.df[y][df_null[x] == False]
        df_null['is_null'] = df_null[y]
        null_index = df_null[df_null['is_null'] == True].index
        x_news = df_x.iloc[null_index]
        x_cubic = df_x.drop(null_index, axis=0)
        y_cubic = df_y.drop(null_index, axis=0)
        f = interp1d(x_cubic, y_cubic, kind=kind)
        self.cubic_dict[(x, y)] = {
            'f': f, 'null_index': null_index, 'x_news': x_news}

    def plot(self, x, y, f, null_index):
        # 画图
        fig, ax = plt.subplots(len(self.cubic_dict.keys()),
                               1, sharex=True, figsize=(6, 6))
        num = 0
        for i, j in self.cubic_dict.items():
            x = i[0]
            y = i[1]
            f = j['f']
            null_index = j['null_index']
            x_plot = np.linspace(self.df[x].min(), self.df[x].max(), 500)
            ax[num].plot(x_plot, f(x_plot))
            ax[num].spines['top'].set_visible(False)
            ax[num].spines['right'].set_visible(False)
            ax[num].set_xlabel(f"{x}")
            ax[num].set_ylabel(f"{y}")
            for i, j in self.df[[x, y]].iloc[null_index].iterrows():
                ax[num].text(j[0], j[1], '%.2f' % j[1])
                ax[num].plot(j[0], j[1], marker=".", markersize=8)
            num += 1
        plt.savefig(f"{x}-{y}.png", dpi=500)

    def cubic_all(self):
        '''
        Func:填充缺失值
        '''
        for i, j in self.cubic_dict.items():
            self.df[i[1]].iloc[j['null_index']] = j['f'](j['x_news'])
        self.plot(x=i[0], y=i[1], f=j['f'], null_index=j['null_index'])

    def missing_summary(self):
        '''
        Func:缺失值所在列
        '''
        return self.df.isnull().any()


if __name__ == '__main__':
    test_dict = {'A': [9, 3, 2, 4, 5, None, 7, 8],
                 'B': [2, 4, 3, None, 1, 13, 14, 12],
                 'C': [3, 5, 7, 8, 9, None, 11, 12],
                 'D': [2, 6, 7, 9, 1, 11, 13, 12]
                 }
    df = pd.DataFrame(test_dict)
    dp = DataPreprocessor(df)
    #print(dp.correlation(columns=['A', 'B', 'C', 'D']))
    dp.cubic(x='B', y='A')
    dp.cubic(x='A', y='B')
    dp.cubic(x='D', y='C')
    dp.cubic_all()
    print(dp.df)
